{
*****************************************************************************
*  Name    : USART2 Library                                                 *
*  Author  : David John Barker                                              *
*  Notice  : Copyright (c) 2003 Mecanique                                   *
*          : All Rights Reserved                                            *
*  Date    : 21/01/06                                                       *
*  Version : 1.6                                                            *
*          : modified SetBaudrate for 16-bit BRG                            *
*          : remove inline asm in WriteItem(pText as string)                *
*  Version : 1.5 Reworked for silicon error                                 *
*          : 1.4 Removed TXIF from WriteItem(Str) and replaced with PIR1, 4 *
*          : 1.3 Added support for USART_BRG16                              *
*  Notes   : 1.2 Supports second USART module                               *
*****************************************************************************
}
module USART2

include "system.bas"

// calculate clock...
const OSC = _clock
const FOSC = OSC * 1000000

// low speed or high speed (BRGH)...
#if IsOption(USART2_BRGH) and not (USART2_BRGH in (true, false))
   #error USART2_BRGH, "Invalid option. BRGH must be TRUE or FALSE."
#endif
#option USART2_BRGH = true
#if USART2_BRGH = true
const USART_MODE = $24
#else
const USART_MODE = $20
#endif

// 8 bit or 16 bit (BRG16 only supported with EUSART modules)...
#if IsOption(USART2_BRG16) and not (USART2_BRG16 in (true, false))
   #error USART2_BRG16, "Invalid option. BRG16 must be TRUE or FALSE."
#endif
#option USART2_BRG16 = false
#if not USART2_BRG16 and not USART2_BRGH     // BRG16 = 0, BRGH = 0
const FMULT = 64
#elseif USART2_BRG16 and not USART2_BRGH     // BRG16 = 1, BRGH = 0
const FMULT = 16
#elseif not USART2_BRG16 and USART2_BRGH     // BRG16 = 0, BRGH = 1
const FMULT = 16
#else                                      // BRG16 = 1, BRGH = 1
const FMULT = 4
#endif

// map registers to USART(x)
#if _usart < 2
   #error _device + " does not support second USART"

// MCU has more than one USART...
#else
public dim                         // -> USART2
   SPBRG_Reg as SPBRG2,
   #if USART2_BRG16
   SPBRGH_Reg as SPBRGH2, // added to support 16-bit BRG for all chips
   BRG16 as BAUDCON2.3,            //    as BAUDCON2
   #endif
   RCRegister as RCREG2,           //    as RCREG2
   TXRegister as TXREG2,           //    as TXREG2
   RCStatus as RCSTA2,             //    as TXSTA2
   TXStatus as TXSTA2,             //    as TXSTA2
 #if _device in (18F24K22, 18F25K22, 18F26K22)
   RCInput As TRISB.Booleans(7),   //    as TRISG (2)
   TXInput As TRISB.Booleans(6)    //    as TRISG (1)   
 #elseif _device in (1846K22)
   RCInput As TRISD.Booleans(7),   //    as TRISG (2)
   TXInput As TRISD.Booleans(6)    //    as TRISG (1)   
 #else
   #warning "Verify TX2/RX2 pin locations"
   RCInput As TRISG.Booleans(2),   //    as TRISG (2)
   TXInput As TRISG.Booleans(1)    //    as TRISG (1)
 #endif
#endif

dim                                // -> USART2
   PIR as PIR3,                    //    as PIR3
   PIE as PIE3,                    //    as PIE3
   IPR as IPR3                     //    as IPR3

// public baudrate constants...
#if USART2_BRG16
public const
   br300 as word   = FOSC / (FMULT * (300 + 1)) - 1 + 0.5,
   br600 as word   = FOSC / (FMULT * (600 + 1)) - 1 + 0.5,
   br1200 as word  = FOSC / (FMULT * (1200 + 1)) - 1 + 0.5,
   br2400 as word  = FOSC / (FMULT * (2400 + 1)) - 1 + 0.5,
   br4800 as word  = FOSC / (FMULT * (4800 + 1)) - 1 + 0.5,
   br9600 as word  = FOSC / (FMULT * (9600 + 1)) - 1 + 0.5,
   br19200 as word = FOSC / (FMULT * (19200 + 1)) - 1 + 0.5,
   br38400 as word = FOSC / (FMULT * (38400 + 1)) - 1 + 0.5,
   br57600 as word = FOSC / (FMULT * (57600 + 1)) - 1 + 0.5,
   br115200 as word = FOSC / (FMULT * (115200 + 1)) - 1 + 0.5
#else
public const
   br300 as byte   = FOSC / (FMULT * (300 + 1)) - 1 + 0.5,
   br600 as byte   = FOSC / (FMULT * (600 + 1)) - 1 + 0.5,
   br1200 as byte  = FOSC / (FMULT * (1200 + 1)) - 1 + 0.5,
   br2400 as byte  = FOSC / (FMULT * (2400 + 1)) - 1 + 0.5,
   br4800 as byte  = FOSC / (FMULT * (4800 + 1)) - 1 + 0.5,
   br9600 as byte  = FOSC / (FMULT * (9600 + 1)) - 1 + 0.5,
   br19200 as byte = FOSC / (FMULT * (19200 + 1)) - 1 + 0.5,
   br38400 as byte = FOSC / (FMULT * (38400 + 1)) - 1 + 0.5,
   br57600 as byte = FOSC / (FMULT * (57600 + 1)) - 1 + 0.5,
   br115200 as byte = FOSC / (FMULT * (115200 + 1)) - 1 + 0.5
#endif

// alias public bitnames to TXSTA(x)...
public dim
   CSRC as TXStatus.7,
   TX9 as TXStatus.6,
   TXEN as TXStatus.5,
   SYNC as TXStatus.4,
   BRGH as TXStatus.2,
   TRMT as TXStatus.1,
   TX9D as TXStatus.0

// alias public bitnames to RCSTA(x)...
public dim
   SPEN as RCStatus.7,
   RX9 as RCStatus.6,
   SREN as RCStatus.5,
   CREN as RCStatus.4,
   ADDEN as RCStatus.3,
   FERR as RCStatus.2,
   OERR as RCStatus.1,
   RX9D as RCStatus.0

// alias public interrupt flags...
public dim
   RCIF as PIR.5,  // receive buffer full
   TXIF as PIR.4,  // transmit buffer empty
   RCIE as PIE.5,  // receive interrupt enable
   TXIE as PIE.4,  // transmit interrupt enable
   RCIP as IPR.5,  // receive interrupt priority
   TXIP as IPR.4   // transmit interrupt priority

// public boolean flags...
public dim
   DataAvailable as PIR.Booleans(5),         // RCIF
   ReadyToSend as PIR.Booleans(4),           // TXIF
   ContinousReceive as RCStatus.Booleans(4), // CREN
   Overrun as RCStatus.Booleans(1),          // OERR
   FrameError as RCStatus.Booleans(2),       // FERR
   RCIEnable as PIE.Booleans(5),             // RCIE
   TXIEnable as PIE.Booleans(4),             // TXIE
   RCIPHigh as IPR.Booleans(5),              // RCIP
   TXIPHigh as IPR.Booleans(4)               // TXIP

public dim
   ReadTerminator as char                    // read string terminator
{
****************************************************************************
* Name    : SetBaudrate                                                    *
* Purpose : Sets the hardware USART baudrate                               *
*         : Pass SPBRG constant, as defined above. For example, br115200   *
****************************************************************************
}
// change parameter to a word... was passed in SPBRGRegister, assuming SPBRG
// and SPBRGH are in contiguous locations but they're not in all chips
public sub SetBaudrate(pSPBRG as word = br19200)
   SPBRG_Reg = pSPBRG.byte0
   #if USART2_BRG16
   SPBRGH_Reg = pSPBRG.byte1
   #endif

   RCStatus = $90         // serial port enable, continuous receive
   RCInput = true         // receive pin is input
   TXInput = false        // transmit pin is output
   TXStatus = USART_MODE  // high or low speed
   #if USART2_BRG16
   BRG16 = 1
   #endif
end sub
{
****************************************************************************
* Name    : ClearOverrun                                                   *
* Purpose : Clear USART overrun error by resetting receive flag            *
****************************************************************************
}
public sub ClearOverrun()
   if Overrun then
      CREN = 0  // disable continuous receive
      CREN = 1  // enable continous receive
   endif
end sub
{
****************************************************************************
* Name    : ReadByte                                                       *
* Purpose : Read a byte from the hardware USART                            *
*         : Waits for USART data, return byte in RCREG(x)                  *
****************************************************************************
}
public function ReadByte() as RCRegister
   repeat
      ClrWDT
   until DataAvailable
end function
{
****************************************************************************
* Name    : ReadBoolean                                                    *
* Purpose : Read a boolean from the hardware USART                         *
****************************************************************************
}
public function ReadBoolean() as boolean
   Result = boolean(ReadByte)
end function
{
****************************************************************************
* Name    : ReadWord                                                       *
* Purpose : Read a word from the hardware USART                            *
****************************************************************************
}
public function ReadWord() as word
   Result.Bytes(0) = ReadByte
   Result.Bytes(1) = ReadByte
end function
{
****************************************************************************
* Name    : ReadLongWord                                                   *
* Purpose : Read a long word from the hardware USART                       *
****************************************************************************
}
public function ReadLongWord() as longword
   Result.Bytes(0) = ReadByte
   Result.Bytes(1) = ReadByte
   Result.Bytes(2) = ReadByte
   Result.Bytes(3) = ReadByte
end function
{
****************************************************************************
* Name    : ReadFloat                                                      *
* Purpose : Read a float from the hardware USART                           *
****************************************************************************
}
public function ReadFloat() as float
   Result.Bytes(0) = ReadByte
   Result.Bytes(1) = ReadByte
   Result.Bytes(2) = ReadByte
   Result.Bytes(3) = ReadByte
end function
{
****************************************************************************
* Name    : WriteByte                                                      *
* Purpose : Write a byte value to the hardware USART                       *
*         : Wait until ready to send is enabled, then send WREG byte       *
****************************************************************************
}
public sub WriteByte(pValue as WREG)
   repeat
      ClrWDT
   until ReadyToSend
   TXRegister = WREG
end sub
{
****************************************************************************
* Name    : WriteBoolean                                                   *
* Purpose : Write a boolean value to the hardware USART                    *
****************************************************************************
}
public sub WriteBoolean(pValue as boolean)
   WriteByte(byte(pValue))
end sub
{
****************************************************************************
* Name    : WriteWord                                                      *
* Purpose : Write a word value to the hardware USART                       *
****************************************************************************
}
public sub WriteWord(pValue as word)
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
end sub
{
****************************************************************************
* Name    : WriteLongWord                                                  *
* Purpose : Write a word value to the hardware USART                       *
****************************************************************************
}
public sub WriteLongWord(pValue as longword)
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
   WriteByte(pValue.Bytes(2))
   WriteByte(pValue.Bytes(3))
end sub
{
****************************************************************************
* Name    : WriteFloat                                                     *
* Purpose : Write a floating point number to the hardware USART            *
****************************************************************************
}
public sub WriteFloat(pValue as float)
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
   WriteByte(pValue.Bytes(2))
   WriteByte(pValue.Bytes(3))
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a boolean from the hardware USART                         *
****************************************************************************
}
sub ReadItem(byref pValue as boolean)
   pValue = boolean(ReadByte)
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a byte from the hardware USART                            *
****************************************************************************
}
sub ReadItem(byref pValue as byte)
   pValue = ReadByte
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a shortint from the hardware USART                        *
****************************************************************************
}
sub ReadItem(byref pValue as shortint)
   pValue = ReadByte
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a word from the hardware USART                            *
****************************************************************************
}
sub ReadItem(byref pValue as word)
   pValue = ReadWord
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read an integer from the hardware USART                        *
****************************************************************************
}
sub ReadItem(byref pValue as integer)
   pValue = ReadWord
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a long word from the hardware USART                       *
****************************************************************************
}
sub ReadItem(byref pValue as longword)
   pValue = ReadLongWord
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a long integer from the hardware USART                    *
****************************************************************************
}
sub ReadItem(byref pValue as longint)
   pValue = ReadLongWord
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a floating point number from the hardware USART           *
****************************************************************************
}
sub ReadItem(byref pValue as float)
   pValue = ReadFloat
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a string from the hardware USART. Use ReadTerminator      *
*         : to specify the input string terminator character.              *
*         : 1.5 Reworked for silicon error                                 *
****************************************************************************
}
sub ReadItem(byref pText as string)
   dim TextPtr as POSTINC0
   dim TextAddr as FSR0
   dim Value as byte

   TextAddr = addressof(pText)
   Value = ReadByte
   while Value <> byte(ReadTerminator)
      TextPtr = Value
      Value = ReadByte
   wend
   TextPtr = null
end sub
{
****************************************************************************
* Name    : Read (COMPOUND)                                                *
* Purpose : Read an item from the hardware USART                           *
****************************************************************************
}
public compound sub Read(ReadItem)
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a boolean value to the hardware USART                    *
****************************************************************************
}
sub WriteItem(pValue as boolean)
   WriteByte(byte(pValue))
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a byte value to the hardware USART                       *
****************************************************************************
}
sub WriteItem(pValue as WREG)
   WriteByte(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a short int value to the hardware USART                  *
****************************************************************************
}
sub WriteItem(pValue as shortint)
   WriteByte(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a word value to the hardware USART                       *
****************************************************************************
}
sub WriteItem(pValue as word)
   WriteWord(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write an integer value to the hardware USART                   *
****************************************************************************
}
sub WriteItem(pValue as integer)
   WriteWord(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a long word to the hardware USART                        *
****************************************************************************
}
sub WriteItem(pValue as longword)
   WriteLongWord(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a long integer to the hardware USART                     *
****************************************************************************
}
sub WriteItem(pValue as longint)
   WriteLongWord(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a floating point number to the hardware USART            *
****************************************************************************
}
sub WriteItem(pValue as float)
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
   WriteByte(pValue.Bytes(2))
   WriteByte(pValue.Bytes(3))
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a string value to the hardware USART                     *
****************************************************************************
}
// jm - remove inline asm
sub WriteItem(pText as string)
   FSR0 = addressof(pText)
    while (INDF0 <> 0)
        WriteByte(POSTINC0)
    end while
end sub
{
****************************************************************************
* Name    : Write (COMPOUND)                                               *
* Purpose : Write an item to the hardware USART                            *
****************************************************************************
}
public compound sub Write(WriteItem)
{
****************************************************************************
* Name    : WaitFor                                                        *
* Purpose : Wait for a byte value to be received                           *
****************************************************************************
}
public function WaitFor(pValue as byte) as boolean
   result = ReadByte = pValue
end function
{
****************************************************************************
* Name    : WaitForTimeout                                                 *
* Purpose : Wait for a byte value to be received, with timeout value in    *
*         : milliseconds                                                   *
****************************************************************************
}
public function WaitForTimeout(pValue as byte, pTimeout as word) as boolean
   dim Counter as byte
   dim Timeout as word

   ClearOverrun
   Result = false
   Counter = 10
   repeat
      Timeout = pTimeout
      while Timeout > 0
         if DataAvailable and (RCRegister = pValue) then
            Result = true
            exit
         endif
         delayus(100)
         dec(Timeout)
      wend
      dec(Counter)
   until Counter = 0
end function
{
****************************************************************************
* Name    : WaitForCount                                                   *
* Purpose : Wait for pAmount bytes to be received. The incoming data is    *
*         : stored in pArray                                               *
****************************************************************************
}
public sub WaitForCount(byref pArray() as byte, pCount as word)
   dim Index as word
   ClearOverrun
   Index = 0
   while pCount > 0
      pArray(Index) = ReadByte
      dec(pCount)
      inc(Index)
   wend
end sub
{
****************************************************************************
* Name    : WaitForStr                                                     *
* Purpose : Wait for a string to be received                               *
****************************************************************************
}
public sub WaitForStr(pStr as string)
   dim Index as byte

   ClearOverrun
   Index = 0
   while pStr(Index) <> null
      if WaitFor(pStr(Index)) then
         inc(Index)
      else
         Index = 0
      endif
   wend
end sub
{
****************************************************************************
* Name    : WaitForStrCount                                                *
* Purpose : Wait for pAmount characters to be received. The incoming data  *
*         : is stored in pStr                                              *
****************************************************************************
}
public sub WaitForStrCount(byref pStr as string, pCount as word)
   dim Index as byte
   ClearOverrun
   Index = 0
   while pCount > 0
      pStr(Index) = ReadByte
      dec(pCount)
      inc(Index)
   wend
   pStr(Index) = null
end sub
{
****************************************************************************
* Name    : WaitForStrTimeout                                              *
* Purpose : Wait for a string to be received, with timeout value in        *
*         : milliseconds                                                   *
****************************************************************************
}
public function WaitForStrTimeout(pStr as string, pTimeout as word) as boolean
   dim Counter, StrIndex as byte
   dim Timeout as word
   dim Ch as byte

   ClearOverrun
   Result = false
   StrIndex = 0
   Counter = 10
   repeat
      Timeout = pTimeout
      while Timeout > 0
         Ch = pStr(StrIndex)
         if Ch = 0 then
            result = true
            exit
         elseif DataAvailable then
            if Ch = RCRegister then
               inc(StrIndex)
            else
               StrIndex = 0
            endif
         endif
         delayus(100)
         dec(Timeout)
      wend
      dec(Counter)
   until Counter = 0
end function
{
****************************************************************************
* Name    : DataAvailableTimeout                                           *
* Purpose : Checks to see if a byte value has been received, with          *
*         : timeout in milliseconds                                        *
****************************************************************************
}
public function DataAvailableTimeout(pTimeout as word) as boolean
   dim Counter as byte
   dim Timeout as word

   ClearOverrun
   Counter = 10
   repeat
      Timeout = pTimeout
      while (Timeout > 0) and not DataAvailable
         delayus(100)
         dec(Timeout)
      wend
      dec(Counter)
   until (Counter = 0) or DataAvailable
   result = DataAvailable
end function
{
****************************************************************************
* Name    : Rep                                                            *
* Purpose : Write a byte value, pAmount times                              *
****************************************************************************
}
public sub Rep(pValue, pAmount as byte)
   while pAmount > 0
      WriteByte(pValue)
      dec(pAmount)
   wend
end sub
{
****************************************************************************
* Name    : Skip                                                           *
* Purpose : Read pAmount bytes                                             *
****************************************************************************
}
public sub Skip(pAmount as byte)
   dim Value as byte
   while pAmount > 0
      Value = ReadByte
      dec(pAmount)
   wend
end sub

// module initialisation...
ReadTerminator = null
